unit spectrum;

{' /*******************************************************************************
'   modSpectrum.pas within vbSpec.vbp
'
'   Main form and ist routines:
'   -Menu handling
'   -FileOpen/Save dialogs
'   -Calls emulator setup dialogs in frmOption/frmDisplayOpt
'   -Run z80 emulation on TApplication.OdIdle- handler
'   -Keyboard handling
'   -Debugging outputs to DelphiSpecDebug.txt
'   -Form painting indirectly via forceRepaint, painting happens in screenPaint
'
'   Original author (vbSpec): Chris Cowley <ccowley@grok.co.uk>
'   Copyright (C)1999-2000 Grok Developments Ltd.
'   http://www.grok.co.uk/
'
'   Translation to Delphi Object Pascal by
'   Jari Korhonen <jarit.korhonen@luukku.com>
'   Copyright (C) 2001- Jari Korhonen
'
'   This program is free software; you can redistribute it and/or
'   modify it under the terms of the GNU General Public License
'   as published by the Free Software Foundation; either version 2
'   of the License, or (at your option) any later version.
'   This program is distributed in the hope that it will be useful,
'   but WITHOUT ANY WARRANTY; without even the implied warranty of
'   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'   GNU General Public License for more details.
'
'   You should have received a copy of the GNU General Public License
'   along with this program; if not, write to the Free Software
'   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
'
' *******************************************************************************/}


interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, Buttons, Menus, ComCtrls, DIB;

type
  TfrmMainWnd = class(TForm)
    mnuMain: TMainMenu;
    mnuFile: TMenuItem;
    mnuOptions: TMenuItem;
    mnuHelp: TMenuItem;
    mnuFileOpen: TMenuItem;
    mnuFileSaveAs: TMenuItem;
    N1: TMenuItem;
    mnuFileResetSpectrum: TMenuItem;
    N2: TMenuItem;
    mnuFileExit: TMenuItem;
    mnuOptionsGeneralSettings: TMenuItem;
    mnuOptionsDisplay: TMenuItem;
    mnuHelpSpectrumKeyboard: TMenuItem;
    mnuHelpAbout: TMenuItem;
    dlgOpen: TOpenDialog;
    statusBar: TStatusBar;
    dlgSave: TSaveDialog;
    DXPaintBoxSpecScreen: TDXPaintBox;
    procedure FormCreate(Sender: TObject);
    procedure MyIdleHandler(Sender: TObject; var Done: Boolean);
    procedure ContinueEmulation(Sender: TObject);
    procedure StopEmulation(Sender: TObject);
    procedure mnuFileOpenClick(Sender: TObject);
    procedure mnuFileExitClick(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure mnuFileResetSpectrumClick(Sender: TObject);
    procedure mnuOptionsGeneralSettingsClick(Sender: TObject);
    procedure mnuOptionsDisplayClick(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure mnuHelpSpectrumKeyboardClick(Sender: TObject);
    procedure mnuHelpAboutClick(Sender: TObject);
    procedure mnuFileSaveAsClick(Sender: TObject);
  private
    procedure WMEraseBkgnd(var m: TWMEraseBkgnd); message WM_ERASEBKGND;
    procedure doKey(down : Boolean; ascii : Word; mods : TShiftState);
    procedure FileOpenDialog;
    procedure FileSaveAsDialog;
    { Private declarations }
  public
    { Public declarations }
    borderColor : byte;
  end;

var
  frmMainWnd: TfrmMainWnd;
  Global_TStates: integer;
  debugFile: TextFile;
  z80runstate: integer;

{$IFDEF USE_DEBUGGING}
procedure ReportDebug(Desc : string; NextPc: integer; forceOutput: boolean);
{$ENDIF}

implementation

uses modMain, modz80, modAY8912, modspectrum, modwaveout, mmsystem, options,
  displayopt, keyboard, about;

{$R *.DFM}


{$IFDEF USE_DEBUGGING}
var ReportDebugCallCtr:integer;

//Print out emulator internal data structures to debug file
//'DelphiSpecDebug.txt'
procedure ReportDebug(Desc : string; NextPc: integer; forceOutput: boolean);
  var Flags: integer; Iff1: integer; Iff2:integer; ii: integer; jj: integer;
      Crcs: array[0..11] of integer;
begin
    if (ReportDebugCallCtr >= 0) and (ReportDebugCallCtr <= 1000000) then
    begin
      if ((ReportDebugCallCtr mod 10000) = 0) or forceOutput=True then
      begin
        Writeln(DebugFile, '---------------------');
        Writeln(DebugFile, Format('%d:%s;%x;%d',
            [ReportDebugCallCtr, Desc, NextPc, Global_TStates]));

        Writeln(DebugFile, 'Registers start');
        Writeln(DebugFile, Format('%s:%x;%x;%x;%x;%x;%x;%x;%x;%x;%x',
          ['A-IX',
          regA,
          regHL,
          regB,
          regC,
          regDE,
          regAF_,
          regHL_,
          regBC_,
          regDE_,
          regIX]));

        Flags := 0;
        if fc then Flags := Flags + F_C;
        if fN then Flags := Flags + F_N;
        if fPV then Flags := Flags + F_PV;
        if f3 then Flags := Flags + F_3;
        if fH then Flags := Flags + F_H;
        if f5 then Flags := Flags + F_5;
        if fZ then Flags := Flags + F_Z;
        if fS then Flags := Flags + F_S;

        if intIff1 then Iff1 := 1 else Iff1:=0;
        if intIff2 then Iff2 := 1 else Iff2:=0;
        Writeln(DebugFile, Format('%s:%x;%x;%x;%x;%x;%x;%x;%x;%x;%x;%x',
          ['IY-Flags',
          regIY,
          regID,
          regSP,
          regPC,
          intI,
          intR,
          intRTemp,
          Iff1,
          Iff2,
          intIM,
          Flags]));
        Writeln(DebugFile, 'Registers end');

        For ii := 0 To 11 do
          Crcs[ii] := ramCrc(ii);

        Writeln(DebugFile, 'RAM Crc start');

        Writeln(DebugFile, Format('%x;%x;%x;%x;%x;%x;%x;%x;%x;%x;%x;%x',
          [Crcs[0],
          Crcs[1],
          Crcs[2],
          Crcs[3],
          Crcs[4],
          Crcs[5],
          Crcs[6],
          Crcs[7],
          Crcs[8],
          Crcs[9],
          Crcs[10],
          Crcs[11]]));
        Writeln(DebugFile);
        Writeln(DebugFile, 'RAM Crc end');

        {Writeln(DebugFile, 'RAM contents start');
        for ii := 0 to 11 do
        begin
          Writeln(DebugFile, Format('%s%d',
            ['RAM Bank: ',
              ii]));
          for jj := 0 to 16383 do
          begin;
            if (jj mod 32) = 0 then
              Writeln(DebugFile);
             Write(DebugFile, Format('%x', [gRamPage[ii, jj]]));
          end;
          Writeln(DebugFile);
        end;
        Writeln(DebugFile, 'RAM contents end');}

        Writeln(DebugFile, Format('%s:%x;%x;%3.5f;%x;%x;%x;%x;%x',
          ['AYRegs1',
          AYPSG.sampleRate,
          AYPSG.register_latch,
          AYPSG.UpdateStep,
          AYPSG.PeriodA,
          AYPSG.PeriodB,
          AYPSG.PeriodC,
          AYPSG.PeriodN,
          AYPSG.PeriodE]));

        Writeln(DebugFile,  Format('%s:%x;%x;%x;%x;%x;%x;%x;%x',
          ['AYRegs2',
          AYPSG.CountA,
          AYPSG.CountB,
          AYPSG.CountC,
          AYPSG.CountN,
          AYPSG.CountE,
          AYPSG.VolA,
          AYPSG.VolB,
          AYPSG.VolC]));

        Writeln(DebugFile, Format('%s:%x;%x;%x;%x;%x;%x;%x;%x',
          ['AYRegs3',
          AYPSG.VolE,
          AYPSG.EnvelopeA,
          AYPSG.EnvelopeB,
          AYPSG.EnvelopeC,
          AYPSG.OutputA,
          AYPSG.OutputB,
          AYPSG.OutputC,
          AYPSG.OutputN]));

        Writeln(DebugFile, Format('%s:%x;%x;%x;%x;%x',
          ['AYRegs4',
          AYPSG.CountEnv,
          AYPSG.Hold,
          AYPSG.Alternate,
          AYPSG.Attack,
          AYPSG.Holding]));

        Writeln(DebugFile, Format('%s:%x;%x;%x;%x;%x;%x;%x;%x',
          ['AYRegsR1',
          AYPSG.Regs[0],
          AYPSG.Regs[1],
          AYPSG.Regs[2],
          AYPSG.Regs[3],
          AYPSG.Regs[4],
          AYPSG.Regs[5],
          AYPSG.Regs[6],
          AYPSG.Regs[7]]));

        Writeln(DebugFile, Format('%s:%x;%x;%x;%x;%x;%x;%x;%x',
          ['AYRegsR2',
          AYPSG.Regs[8],
          AYPSG.Regs[9],
          AYPSG.Regs[10],
          AYPSG.Regs[11],
          AYPSG.Regs[12],
          AYPSG.Regs[13],
          AYPSG.Regs[14],
          AYPSG.Regs[15]]));

        For ii := 0 To 7 do
          Writeln(DebugFile,  Format('%s:%x;%x;%x;%x;%x;%x;%x;%x',
            ['AYRegsV1',
            AYPSG.VolTable2[8 * ii],
            AYPSG.VolTable2[8 * ii + 1],
            AYPSG.VolTable2[8 * ii + 2],
            AYPSG.VolTable2[8 * ii + 3],
            AYPSG.VolTable2[8 * ii + 4],
            AYPSG.VolTable2[8 * ii + 5],
            AYPSG.VolTable2[8 * ii + 6],
            AYPSG.VolTable2[8 * ii + 7]]));
      end;
    end;
    Inc(ReportDebugCallCtr);
end;
{$ENDIF}


//Don't let Windows erase the background of the window
//because we fill everything ourselves
procedure TfrmMainWnd.WMEraseBkgnd(var m : TWMEraseBkgnd);
begin
  m.Result := LRESULT(False);
end;


//This is DelphiSpec's "main loop"
//It runs spectrum for one interrupt period (50 ms spectrum time)
//and tells then Windows it needs more processor cycles (Done:=False).
//This way it is a definite CPU hog, but should be fast too
//During setting screens, emulation is paused (z80runstate=Z80PAUSED)
procedure TfrmMainWnd.MyIdleHandler(Sender: TObject; var Done: Boolean);
begin
  if (z80runstate=Z80RUNNING) then
  begin
    //Execute some z80 code
    execute(Global_TStates);
    Done := false;
  end
  else
  begin
    Sleep(100);
    Done := true;
  end;
end;

//When app gains focus
procedure TfrmMainWnd.ContinueEmulation(Sender: TObject);
begin;
  if z80runstate=Z80STOPPED then
    z80runstate := Z80RUNNING;
end;


//When app loses focus
procedure TfrmMainWnd.StopEmulation(Sender: TObject);
begin;
  if z80runstate=Z80RUNNING then
    z80runstate := Z80STOPPED;
end;


procedure TfrmMainWnd.FormCreate(Sender: TObject);
  var ThreadHandle: THandle;
begin
{$IFDEF USE_DEBUGGING}
  AssignFile(debugFile, 'DelphiSpecDebug.txt');
  Rewrite(debugFile);
  ReportDebugCallCtr:=0;
{$ENDIF}
  ThreadHandle := GetCurrentThread;
  SetThreadPriority(ThreadHandle, THREAD_PRIORITY_HIGHEST);
  z80runstate := Z80RUNNING;
  Application.OnIdle:= MyIdleHandler;
  Application.OnActivate:=ContinueEmulation;
  Application.OnDeactivate:=StopEmulation;
  Main;
  Global_TStates := -glTstatesPerInterrupt;

//  LoadSnaSnap(GetAppPath + '\manic.sna');
//  SaveZ80Snap(GetAppPath + '\mymanic.z80');
end;

procedure TfrmMainWnd.FileOpenDialog;
  var Ext: string;
begin
    if (dlgOpen.execute) then
    begin
      If dlgOpen.FileName <> '' Then
      begin
        Ext := LowerCase(ExtractFileExt(dlgOpen.FileName));

        if (Ext = '.z80') then
          LoadZ80Snap(dlgOpen.FileName)
        else if (Ext = '.rom') then
        begin
          Z80Reset;
          SetEmulatedModel(0);
          if LoadROM(dlgOpen.FileName) = false then
            Application.Terminate;
        end
        else if (Ext = '.sna') then
          LoadSNASnap(dlgOpen.FileName)
        else
        begin
          //try opening it as a SNA file
          LoadSNASnap(dlgOpen.FileName);
        end;
      end;
    end;
End;


procedure TfrmMainWnd.mnuFileOpenClick(Sender: TObject);
  var ii: integer;
begin
  {$IFDEF USE_SOUND}
      If gbSoundEnabled Then waveOutReset(glphWaveOut);
  {$ENDIF}

  FileOpenDialog;
  forceRepaint;
  {$IFDEF USE_SOUND}
      If gbSoundEnabled Then
      begin
          waveOutReset(glphWaveOut);
          For ii := 1 To 12 do
          begin
              waveOutWrite(glphWaveOut,
                @gtWavHdr[ii], sizeof(gtWavHdr[ii]));
          end;
          glBufNum := 12;
      End;
  {$ENDIF}
end;

procedure TfrmMainWnd.mnuFileExitClick(Sender: TObject);
begin
    {$IFDEF USE_SOUND}
    If gbSoundEnabled Then CloseWaveOut;
    {$ENDIF}

    timeEndPeriod(1);
    Application.Terminate;
end;

procedure TfrmMainWnd.doKey(down : Boolean; ascii : Word; mods : TShiftState);
    var CAPS : Boolean; SYMB : Boolean;
    //Shift : Boolean;
begin

    CAPS := (ssShift in mods);
    SYMB := (ssCtrl in mods);

    // Change control versions of keys to lower case
    If (ascii >= 1) And (ascii <= $27) And SYMB Then
        ascii := ascii + Ord('a') - 1;

    If CAPS Then keyCAPS_V := (keyCAPS_V And (Not 1)) Else keyCAPS_V := (keyCAPS_V Or 1);
    If SYMB Then keyB_SPC := (keyB_SPC And (Not 2)) Else keyB_SPC := (keyB_SPC Or 2);

    Case ascii of
      8:  //Backspace
        begin
          If down Then
          begin
            key6_0 := (key6_0 And (Not 1));
            keyCAPS_V := (keyCAPS_V And (Not 1));
          end
          Else
          begin
            key6_0 := (key6_0 Or 1);
            If Not CAPS Then
                keyCAPS_V := (keyCAPS_V Or 1);
          End;
        end;
      65: //A
        If down Then keyA_G := (keyA_G And (Not 1)) Else keyA_G := (keyA_G Or 1);
      66: //B
        If down Then keyB_SPC := (keyB_SPC And (Not 16)) Else keyB_SPC := (keyB_SPC Or 16);
      67: //C
        If down Then keyCAPS_V := (keyCAPS_V And (Not 8)) Else keyCAPS_V := (keyCAPS_V Or 8);
      68: //D
        If down Then keyA_G := (keyA_G And (Not 4)) Else keyA_G := (keyA_G Or 4);
      69: //E
        If down Then keyQ_T := (keyQ_T And (Not 4)) Else keyQ_T := (keyQ_T Or 4);
      70: //F
        If down Then keyA_G := (keyA_G And (Not 8)) Else keyA_G := (keyA_G Or 8);
      71: //G
        If down Then keyA_G := (keyA_G And (Not 16)) Else keyA_G := (keyA_G Or 16);
      72: //H
        If down Then keyH_ENT := (keyH_ENT And (Not 16)) Else keyH_ENT := (keyH_ENT Or 16);
      73: //I
        If down Then keyY_P := (keyY_P And (Not 4)) Else keyY_P := (keyH_ENT Or 4);
      74: //J
        If down Then keyH_ENT := (keyH_ENT And (Not 8)) Else keyH_ENT := (keyH_ENT Or 8);
      75: //K
        If down Then keyH_ENT := (keyH_ENT And (Not 4)) Else keyH_ENT := (keyH_ENT Or 4);
      76: //L
        If down Then keyH_ENT := (keyH_ENT And (Not 2)) Else keyH_ENT := (keyH_ENT Or 2);
      77: //M
        If down Then keyB_SPC := (keyB_SPC And (Not 4)) Else keyB_SPC := (keyB_SPC Or 4);
      78: //N
        If down Then keyB_SPC := (keyB_SPC And (Not 8)) Else keyB_SPC := (keyB_SPC Or 8);
      79: //O
        If down Then keyY_P := (keyY_P And (Not 2)) Else keyY_P := (keyY_P Or 2);
      80: //P
        If down Then keyY_P := (keyY_P And (Not 1)) Else keyY_P := (keyY_P Or 1);
      81: //Q
        If down Then keyQ_T := (keyQ_T And (Not 1)) Else keyQ_T := (keyQ_T Or 1);
      82: //R
        If down Then keyQ_T := (keyQ_T And (Not 8)) Else keyQ_T := (keyQ_T Or 8);
      83: //S
        If down Then keyA_G := (keyA_G And (Not 2)) Else keyA_G := (keyA_G Or 2);
      84: //T
        If down Then keyQ_T := (keyQ_T And (Not 16)) Else keyQ_T := (keyQ_T Or 16);
      85: //U
        If down Then keyY_P := (keyY_P And (Not 8)) Else keyY_P := (keyY_P Or 8);
      86: //V
        If down Then keyCAPS_V := (keyCAPS_V And (Not 16)) Else keyCAPS_V := (keyCAPS_V Or 16);
      87: //W
        If down Then keyQ_T := (keyQ_T And (Not 2)) Else keyQ_T := (keyQ_T Or 2);
      88: //X
        If down Then keyCAPS_V := (keyCAPS_V And (Not 4)) Else keyCAPS_V := (keyCAPS_V Or 4);
      89: //Y
        If down Then keyY_P := (keyY_P And (Not 16)) Else keyY_P := (keyY_P Or 16);
      90: //Z
        If down Then keyCAPS_V := (keyCAPS_V And (Not 2)) Else keyCAPS_V := (keyCAPS_V Or 2);
      48: //0
        If down Then key6_0 := (key6_0 And (Not 1)) Else key6_0 := (key6_0 Or 1);
      49: //1
        If down Then key1_5 := (key1_5 And (Not 1)) Else key1_5 := (key1_5 Or 1);
      50: //2
        If down Then key1_5 := (key1_5 And (Not 2)) Else key1_5 := (key1_5 Or 2);
      51: //3
        If down Then key1_5 := (key1_5 And (Not 4)) Else key1_5 := (key1_5 Or 4);
      52: //4
        If down Then key1_5 := (key1_5 And (Not 8)) Else key1_5 := (key1_5 Or 8);
      53: //5
        If down Then key1_5 := (key1_5 And (Not 16)) Else key1_5 := (key1_5 Or 16);
      54: //6
        If down Then key6_0 := (key6_0 And (Not 16)) Else key6_0 := (key6_0 Or 16);
      55: //7
        If down Then key6_0 := (key6_0 And (Not 8)) Else key6_0 := (key6_0 Or 8);
      56: //8
        If down Then key6_0 := (key6_0 And (Not 4)) Else key6_0 := (key6_0 Or 4);
      57: //9
        If down Then key6_0 := (key6_0 And (Not 2)) Else key6_0 := (key6_0 Or 2);
      96: //Keypad 0
        If down Then key6_0 := (key6_0 And (Not 1)) Else key6_0 := (key6_0 Or 1);
      97: //Keypad 1
        If down Then key1_5 := (key1_5 And (Not 1)) Else key1_5 := (key1_5 Or 1);
      98: //Keypad 2
        If down Then key1_5 := (key1_5 And (Not 2)) Else key1_5 := (key1_5 Or 2);
      99: //Keypad 3
        If down Then key1_5 := (key1_5 And (Not 4)) Else key1_5 := (key1_5 Or 4);
      100:  //Keypad 4
        If down Then key1_5 := (key1_5 And (Not 8)) Else key1_5 := (key1_5 Or 8);
      101:  //Keypad 5
        If down Then key1_5 := (key1_5 And (Not 16)) Else key1_5 := (key1_5 Or 16);
      102:  //Keypad 6
        If down Then key6_0 := (key6_0 And (Not 16)) Else key6_0 := (key6_0 Or 16);
      103:  //Keypad 7
        If down Then key6_0 := (key6_0 And (Not 8)) Else key6_0 := (key6_0 Or 8);
      104:  //Keypad 8
        If down Then key6_0 := (key6_0 And (Not 4)) Else key6_0 := (key6_0 Or 4);
      105:  //Keypad 9
        If down Then key6_0 := (key6_0 And (Not 2)) Else key6_0 := (key6_0 Or 2);
      106:  //Keypad *
        begin
          If down Then
            keyB_SPC := (keyB_SPC And Not (18))
          Else
          begin
            If SYMB Then
                keyB_SPC := (keyB_SPC Or 16)
            Else
                keyB_SPC := (keyB_SPC Or 18);
          End;
        end;
      107:  //Keypad +
        begin
          If down Then
          begin
            keyH_ENT := (keyH_ENT And (Not 4));
            keyB_SPC := (keyB_SPC And (Not 2));
          end
          Else
          begin
            keyH_ENT := (keyH_ENT Or 4);
            If Not SYMB Then
                keyB_SPC := (keyB_SPC Or 2);
          End;
        end;
      109:  //Keypad -
        begin
          If down Then
          begin
            keyH_ENT := (keyH_ENT And (Not 8));
            keyB_SPC := (keyB_SPC And (Not 2));
          end
          Else
          begin
            keyH_ENT := (keyH_ENT Or 8);
            If Not SYMB Then
                keyB_SPC := (keyB_SPC Or 2);
          End;
        end;
      110:  //Keypad .
        begin
          If down Then
            keyB_SPC := (keyB_SPC And (Not 6))
          Else
          begin
            If SYMB Then
                keyB_SPC := (keyB_SPC Or 4)
            Else
                keyB_SPC := (keyB_SPC Or 6);
          End;
        end;
      111:  //Keypad /
        begin
          If down Then
          begin
            keyCAPS_V := (keyCAPS_V And (Not 16));
            keyB_SPC := (keyB_SPC And (Not 2));
          end
          Else
          begin
            keyCAPS_V := (keyCAPS_V Or 16);
            If Not SYMB Then
                keyB_SPC := (keyB_SPC Or 2);
          End;
        end;
      37: //Left
        begin
          If down Then
          begin
            key1_5 := (key1_5 And (Not 16));
            keyCAPS_V := (keyCAPS_V And (Not 1));
          end
          Else
          begin
            key1_5 := (key1_5 Or 16);
            If Not SYMB Then
                keyB_SPC := (keyB_SPC Or 2);
          End;
        end;
      38: //Up
        begin
          If down Then
          begin
            key6_0 := (key6_0 And (Not 8));
            keyCAPS_V := (keyCAPS_V And (Not 1));
          end
          Else
          begin
            key6_0 := (key6_0 Or 8);
            If Not CAPS Then
                keyCAPS_V := (keyCAPS_V Or 1);
          End;
        end;
      39: //Right
        begin
          If down Then
          begin
            key6_0 := (key6_0 And (Not 4));
            keyCAPS_V := (keyCAPS_V And (Not 1));
          end
          Else
          begin
            key6_0 := (key6_0 Or 4);
            If Not CAPS Then
                keyCAPS_V := (keyCAPS_V Or 1);
          End;
        end;
      40: //Down
        begin
          If down Then
          begin
            key6_0 := (key6_0 And (Not 16));
            keyCAPS_V := (keyCAPS_V And (Not 1));
          end
          Else
          begin
            key6_0 := (key6_0 Or 16);
            If Not CAPS Then
                keyCAPS_V := (keyCAPS_V Or 1);
          End;
        end;
      13: //RETURN
        If down Then keyH_ENT := (keyH_ENT And (Not 1)) Else keyH_ENT := (keyH_ENT Or 1);
      32: //SPACE BAR
        If down Then keyB_SPC := (keyB_SPC And (Not 1)) Else keyB_SPC := (keyB_SPC Or 1);
      187:  //=/+ key
        begin
          If down Then
          begin
            If CAPS Then
                keyH_ENT := (keyH_ENT And (Not 4))
            Else
                keyH_ENT := (keyH_ENT And (Not 2));

            keyB_SPC := (keyB_SPC And (Not 2));
            keyCAPS_V := (keyCAPS_V Or 1);
          end
          Else
          begin
            keyH_ENT := (keyH_ENT Or 4);
            keyH_ENT := (keyH_ENT Or 2);
            keyB_SPC := (keyB_SPC Or 2);
          End;
        end;
      189:  // -/_ key
        begin
          If down Then
          begin
            If CAPS Then
                key6_0 := (key6_0 And (Not 1))
            Else
                keyH_ENT := (keyH_ENT And (Not 8));

            keyB_SPC := (keyB_SPC And (Not 2));
            keyCAPS_V := (keyCAPS_V Or 1);
          end
          Else
          begin
            key6_0 := (key6_0 Or 1);        // Release the Spectrum's '0' key
            keyH_ENT := (keyH_ENT Or 8);    // Release the Spectrum's 'J' key
            keyB_SPC := (keyB_SPC Or 2);    // Release the Symbol Shift key
          End;
        end;
      186:  // ;/: keys
        begin
          If down Then
          begin
            If CAPS Then
                keyCAPS_V := (keyCAPS_V And (Not 2))
            Else
                keyY_P := (keyY_P And (Not 2));

            keyB_SPC := (keyB_SPC And (Not 2));
            keyCAPS_V := (keyCAPS_V Or 1);
          end
          Else
          begin
            keyCAPS_V := (keyCAPS_V Or 2);
            keyY_P := (keyY_P Or 2);
            keyB_SPC := (keyB_SPC Or 2);
          End;
        end;
    End;
End;


procedure TfrmMainWnd.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
    If ssAlt in Shift Then Exit;
    doKey(True, Key, Shift);
end;

procedure TfrmMainWnd.FormKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
    doKey(False, Key, Shift);
end;

procedure TfrmMainWnd.mnuFileResetSpectrumClick(Sender: TObject);
  var ii: integer;
begin
  {$IFDEF USE_SOUND}
  If gbSoundEnabled Then
  begin
      waveOutReset(glphWaveOut);
      For ii := 1 To 12 do
          waveOutWrite(glphWaveOut, @gtWavHdr[ii], sizeof(gtWavHdr[ii]));

      glBufNum := 12;
  End;
  {$ENDIF}
  Z80Reset;
  resetKeyboard;
end;

procedure TfrmMainWnd.mnuOptionsGeneralSettingsClick(Sender: TObject);
  var ii: integer;
begin
  z80runstate:=Z80PAUSED;
  frmOptions.Showmodal;
  {$IFDEF USE_SOUND}
  If gbSoundEnabled Then
  begin
      waveOutReset(glphWaveOut);
      For ii := 1 To 12 do
          waveOutWrite(glphWaveOut, @gtWavHdr[ii], sizeof(gtWavHdr[ii]));

      glBufNum := 12;
  end;
  {$ENDIF}
  z80runstate:=Z80RUNNING;
end;

procedure TfrmMainWnd.mnuOptionsDisplayClick(Sender: TObject);
  var ii:integer;
begin
    z80runstate:=Z80PAUSED;
    frmDisplayOpt.ShowModal;

    {$IFDEF USE_SOUND}
    If gbSoundEnabled Then
    begin
        waveOutReset(glphWaveOut);
        For ii := 1 To 12 do
            waveOutWrite(glphWaveOut, @gtWavHdr[glBufNum], sizeof(gtWavHdr[glBufNum]));

        glBufNum := 12;
    End;
    {$ENDIF}
  z80runstate:=Z80RUNNING;
end;

procedure TfrmMainWnd.FormPaint(Sender: TObject);
begin
  forceRepaint;
end;

procedure TfrmMainWnd.mnuHelpSpectrumKeyboardClick(Sender: TObject);
begin
  //Keyboard box is modeless
  frmKeyboard.Show;
end;

procedure TfrmMainWnd.mnuHelpAboutClick(Sender: TObject);
begin
  z80runstate:=Z80PAUSED;
  frmAbout.ShowModal;
  z80runstate:=Z80RUNNING;
end;

procedure TfrmMainWnd.FileSaveAsDialog;
  var Ext: string;
begin
    dlgSave.execute;

    if dlgSave.FileName <> '' Then
    begin
      Ext := LowerCase(ExtractFileExt(dlgSave.FileName));
      
      if (Ext = '.z80') then
        SaveZ80Snap(dlgSave.FileName)
      else if (Ext = '.rom') then
        SaveROM(dlgSave.FileName)
      else if (Ext = '.sna') then
        SaveSNASnap(dlgSave.FileName)
      else  //save it as a Z80 file
        SaveZ80Snap(dlgSave.FileName);
    end;
end;

procedure TfrmMainWnd.mnuFileSaveAsClick(Sender: TObject);
  var ii:integer;
begin
  {$IFDEF USE_SOUND}
  If gbSoundEnabled Then waveOutReset(glphWaveOut);
  {$ENDIF}

  FileSaveAsDialog;

  {$IFDEF USE_SOUND}
  If gbSoundEnabled Then
  begin
      waveOutReset(glphWaveOut);
      For ii := 1 To 12 do
          waveOutWrite(glphWaveOut, @gtWavHdr[glBufNum], sizeof(gtWavHdr[glBufNum]));

      glBufNum := 12;
  End;
  {$ENDIF}
end;

end.


